namespace CSLint
{
    using System;
    using System.Text;
    using Reflector.CodeModel;

    internal class GraphEdge
    {
        internal GraphEdge    next;
        internal object       code;
        internal MethodDesc   methodDesc;
        internal GraphVertex  vertex;
        internal CallContext  callCtx;
        internal long         mask;

        internal GraphEdge(GraphVertex vertex, MethodDesc methodDesc, object code, GraphEdge sibling, CallContext cc)
        {
            next = sibling;
            callCtx = cc;
            mask = 0;
            this.code = code;
            this.vertex = vertex;
            this.methodDesc = methodDesc;
        }
		
        internal void dump(GraphVertex from) 
        {
            Console.Error.WriteLine("Edge from vertex " + from + " to vertex " + vertex + " in method " + methodDesc);
        }

        internal void message(StringBuilder sb, int loopId)
        {
            mask |= (1L << (loopId & 63));
            String msg = "Loop " + loopId + ": lock(" + vertex + ") in method " + 
                    methodDesc + " can cause deadlock";
            String srcFile = methodDesc.cls.getSourceFilePath();
            if (srcFile != null) 
            {
                sb.AppendLine(srcFile + ":" + methodDesc.getLineNumber(code) + ": " + msg);
            } 
            else 
            {
                sb.AppendLine(msg);
            }
            for (CallContext cc = callCtx; cc.inv != null; cc = cc.up)
            {
                msg = "  Loop " + loopId + ": called from " + cc.inv.caller;
                srcFile = cc.inv.caller.cls.getSourceFilePath();
                if (srcFile != null) 
                {
                    sb.AppendLine(srcFile + ":" + cc.inv.caller.getLineNumber(cc.inv.expr) + ": " + msg);
                } 
                else 
                {
                    sb.AppendLine(msg);
                }
            }
        }
    }
}