namespace CSLint
{
    using System;
    using System.Text;
    using Reflector.CodeModel;

    public class GraphVertex
    {
        internal GraphEdge   edges;
        internal GraphVertex next;
        internal GraphVertex pred;
        internal GraphEdge   backEdge;
        internal GraphVertex breadthFirst;
        internal int         marker;
        
        internal Object      syncObject;

        internal static GraphVertex graph;		
		
        public override String ToString() 
        {
            IMemberReference mbr = syncObject as IMemberReference;
            if (mbr != null)
            {
                return ClassDesc.GetTypeName(mbr.DeclaringType as ITypeReference) + "." + mbr.Name;
            }
            ITypeReference type = syncObject as ITypeReference;
            if (type != null)
            {
                return ClassDesc.GetTypeName(type);
            }
            ITypeOfExpression typeOf = syncObject as ITypeOfExpression;
            if (typeOf != null)
            {
                return "typeof(" + ClassDesc.GetTypeName(typeOf.Type as ITypeReference) + ")";
            }
            return syncObject.ToString();
         }

        internal static String verify()
        {
            StringBuilder sb = new StringBuilder();
            int nLoops = 0;
            int marker = 1;
            for (GraphVertex root = graph; root != null; marker++, root = root.next)
            {
                root.breadthFirst = null;
                root.marker = marker;
                bool loopFound = false;
                GraphVertex thisLevel = root;
                do 
                {
                    GraphVertex nextLevel = null;
                    for (GraphVertex v = thisLevel; v != null; v = v.breadthFirst)
                    {
                        for (GraphEdge e = v.edges; e != null; e = e.next)
                        {
                            GraphVertex ev = e.vertex;
                            if (ev == root)
                            {
                                long mask = e.mask;
                                GraphVertex p;
                                for (p = v; p != root; p = p.pred)
                                {
                                    mask &= p.backEdge.mask;
                                }
                                loopFound = true;
                                if (mask == 0)
                                {
                                    nLoops += 1;
                                    e.message(sb, nLoops);
                                    for (p = v; p != root; p = p.pred)
                                    {
                                        p.backEdge.message(sb, nLoops);
                                    }
                                }
                            }
                            else if (ev.marker != marker)
                            {
                                ev.pred = v;
                                ev.marker = marker;
                                ev.breadthFirst = nextLevel;
                                nextLevel = ev;
                                ev.backEdge = e;
                            }
                        }
                    }
                    // Make it possible to GC to collection unused contexts
                    while (thisLevel != null)
                    {
                        GraphVertex nextSibling = thisLevel.breadthFirst;
                        thisLevel.breadthFirst = null;
                        thisLevel = nextSibling;
                    }
                    thisLevel = nextLevel;
                }
                while (thisLevel != null && (!loopFound || CSLintAddin.instance.reportAllLoops));
            }   
            if (nLoops == 0) 
            { 
                sb.AppendLine("No deadlocks found");
            }
            return sb.ToString();
        }
		
        internal void addEdge(GraphVertex vertex, MethodDesc caller, object code, CallContext cc)
        {
            for (GraphEdge edge = edges; edge != null; edge = edge.next)
            {
                if (edge.vertex == vertex && edge.methodDesc == caller && edge.code == code && (!CSLintAddin.instance.reportAllLoops || edge.callCtx.inv == cc.inv))
                {
                    return;
                }
            }
            edges = new GraphEdge(vertex, caller, code, edges, cc);
            //edges.dump(this);
        }
		
        public GraphVertex(object sync)
        {
            syncObject = sync;
            marker = 0;
            next = graph;
            graph = this;
            edges = null;
        }
    }
}