namespace CSLint
{
    using System;
    using Reflector.CodeModel;

    internal class MethodDesc
    {
        internal IMethodDeclaration method;
        internal InvocationDesc invocations;
        internal MethodDesc next;
        internal CallContext proceededStates;
        internal int nStates;

        internal ClassDesc cls;

        internal OverriddenMethod overriddenBy;
        internal ILockStatement[] syncInsns;
        internal GraphVertex[] topLevelSyncBlocks;
        internal MethodDesc nextEntryPoint;


        internal MethodDesc(MethodDesc next, ClassDesc cls, IMethodDeclaration method)
        {
            this.next = next;
            this.cls = cls;
            this.method = method;
        }

        internal void addState(CallContext cc)
        {
            cc.next = proceededStates;
            proceededStates = cc;
            nStates += 1;
        }

        internal void addOverriddenBy(MethodDesc md)
        {
            overriddenBy = new OverriddenMethod(md, overriddenBy);
        }


        internal void addInvocation(MethodDesc method, bool isVirtual, IExpression expr, VertexList monitors)
        {
            invocations = new InvocationDesc(invocations, this, method, expr, isVirtual, monitors);
        }

        public override String ToString()
        {
            return ClassDesc.GetTypeName(method.DeclaringType as ITypeReference) + "." + method.Name;
        }

        internal bool isProceeded(CallContext cc)
        {
            for (CallContext state = proceededStates; state != null; state = state.next)
            {
                if (state.equalState(cc))
                {
                    return true;
                }
            }
            return false;
        }


        internal void addOverriddenMethodInvocations()
        {
            for (InvocationDesc inv = invocations; inv != null; inv = inv.next)
            {
                if (inv.isVirtual)
                {
                    for (OverriddenMethod om = inv.method.overriddenBy; om != null; om = om.next)
                    {
                        invocations = new InvocationDesc(invocations, om.method, inv);
                    }
                }
            }
        }

        internal int getLineNumber(object insn)
        {
            //TODO:  don't know how to extract line number information from CLI
            return 0;
        }


        internal void buildConcurrentClosure()
        {
            int i, j;
            InvocationDesc thisInvocation = new InvocationDesc(this);
            CallContext.entryPoint.inv = thisInvocation;
            CallContext thisLevel = CallContext.entryPoint;
            do
            {
                CallContext nextLevel = null;
                for (CallContext cc = thisLevel; cc != null; cc = cc.breadthFirst)
                {
                    MethodDesc mth = cc.inv.method;
                    if (cc.monitors.Length != 0 && mth.topLevelSyncBlocks != null)
                    {
                        for (i = mth.topLevelSyncBlocks.Length; --i >= 0; )
                        {
                            for (j = cc.monitors.Length; --j >= 0 && mth.topLevelSyncBlocks[i] != cc.monitors[j]; )
                                ;
                            if (j < 0)
                            {
                                cc.monitors[cc.monitors.Length - 1].addEdge(mth.topLevelSyncBlocks[i],
                                                                            mth,
                                                                            mth.syncInsns[i],
                                                                            cc);
                            }
                        }
                    }
                    for (InvocationDesc inv = mth.invocations; inv != null; inv = inv.next)
                    {
                        CallContext cc1 = new CallContext(cc, nextLevel, inv);
                        if (!inv.method.isProceeded(cc1))
                        {
                            inv.method.addState(cc1);
                            nextLevel = cc1;
                        }
                    }
                }
                // Make it possible to GC to collection unused contexts
                while (thisLevel != null)
                {
                    CallContext nextSibling = thisLevel.breadthFirst;
                    thisLevel.breadthFirst = null;
                    thisLevel = nextSibling;
                }
                thisLevel = nextLevel;
            }
            while (thisLevel != null);

            CallContext.entryPoint.inv = null;
        }
    }
}